/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <config.h>

#include <doca_flow.h>
#include <sys/types.h>

#include "doca-mirror.h"
#include "id-fpool.h"
#include "netdev-offload.h"
#include "openvswitch/vlog.h"
#include "openvswitch/list.h"
#include "ovs-doca.h"
#include "refmap.h"
#include "util.h"

/*
 * Mirrros in OVS DOCA OFFLOAD:
 *
 * Upper layer (netdev-dpdk-offload) provides ovs doca offload layer (dpdk-offload-doca) an
 * ordered RTE_FLOW_ACTION list. Mirrors are all the output ports before the last forward
 * action (JUMP, OUTPUT, ...).
 * Upper layer put those mirrored outputs in a OVS_RTE_FLOW_ACTION_TYPE(NESTED) which
 * has a nested sample action with ratio 1 that executes a sample action list with an output.
 *
 * Each such nested action output is taken to a shared resource mirror configuration target.
 * The last forward action is taken to the entry's forward.
 *
 * A mirror shared resource is created per such target configuration/bindable_object,
 * then configured and bound to a given bindable_object, before being attached to a an
 * entry.
 *
 * Example of action list with mirrors:
 * set_ipv4_src ipv4_addr 1.1.1.1 / nested_actions(represented_port ethdev_port_id 3) /
 * nested_actions(represented_port ethdev_port_id 2) / represented_port ethdev_port_id 1 / end
 * In this case, the nested output ports 2 and 3 are taken to a mirror shared resource
 * targets (nr_targets=2) and used as a key to find a similar shared mirror resource.
 * The created/found shared resource is set as the entry's monitor, and output to port 1
 * will be the entry's forward (together with entry's action to set ipv4).
 *
 * Example of action list with mirrors and needing an action split (done by upper layer):
 * nested(out1), nested(out2), push vlan, nested(out3), out4
 *
 * This action list will cause the original rule will be split to two rules:
 *   Rule 1 actions: nested(out1), nested(out2), set_tag, forward to continuation table
 *   Rule 2 actions: push vlan, nested(out3), out4
 *
 * Rule 1 will create a shared mirror with nr_targets=2, targets=1,2
 * Rule 2 will create a shared mirror with nr_targets=1, targets=3
 */

VLOG_DEFINE_THIS_MODULE(doca_mirrors);

static struct ovs_list mirrors_binding_list = OVS_LIST_INITIALIZER(&mirrors_binding_list);
static struct ovs_mutex bound_mirrors_lock = OVS_MUTEX_INITIALIZER;
static atomic_count mirrors_count = ATOMIC_COUNT_INIT(0);

struct doca_mirrors {
    uint32_t base_id;
    struct refmap *rfm;
    struct id_fpool *id_pool;
};

struct doca_mirror_binding {
    void *bindable_object;
    uint32_t shared_mirror_id;
    struct doca_mirror_ctx *mirror_ctx;
    struct ovs_list mirror_ctx_list_node;
    struct ovs_list mirrors_binding_list_node;
};

struct doca_mirror_ctx {
    struct doca_mirrors *doca_mirrors;
    struct ovs_list binding_list;
};

OVS_ASSERT_PACKED(struct doca_mirror_key,
    struct doca_flow_mirror_target targets[DOCA_MIRRORS_MAX_TARGETS];
    char pad[4];
    int nr_targets;
);

struct doca_mirror_arg {
    struct doca_mirrors *doca_mirrors;
    struct doca_mirror_key *key;
};

static int
doca_mirror_ctx_init(void *ctx_, void *arg_)
{
    struct doca_mirror_ctx *mirror_ctx = ctx_;
    struct doca_mirror_arg *arg = arg_;

    mirror_ctx->doca_mirrors = arg->doca_mirrors;
    ovs_list_init(&mirror_ctx->binding_list);

    return 0;
}

static void
doca_mirror_ctx_uninit(void *ctx_ OVS_UNUSED)
{
}

static struct ds *
doca_mirror_ctx_dump(struct ds *s, void *key_, void *ctx_ OVS_UNUSED, void *arg_)
{
    struct doca_mirror_key *key = key_;
    struct doca_mirror_arg *arg = arg_;
    int i;

    ds_put_format(s, "doca_mirrors: %p, nr_targets=%d", arg->doca_mirrors, key->nr_targets);
    for (i = 0; i < key->nr_targets; i++) {
        if (key->targets[i].fwd.type == DOCA_FLOW_FWD_PORT) {
            ds_put_format(s, ", target[%d].fwd.port_id=%d", i, key->targets[i].fwd.port_id);
        } else if (key->targets[i].fwd.type == DOCA_FLOW_FWD_PIPE) {
            ds_put_format(s, ", target[%d].fwd.next_pipe=%p", i, key->targets[i].fwd.next_pipe);
        } else {
            ds_put_format(s, ", target[%d].fwd.type=%d", i, key->targets[i].fwd.type);
        }
    }

    return s;
}

struct doca_mirrors *
doca_mirrors_create(uint32_t esw_id)
{
    struct doca_mirrors *doca_mirrors;

    doca_mirrors = xzalloc(sizeof *doca_mirrors);
    doca_mirrors->id_pool = id_fpool_create(MAX_OFFLOAD_THREAD_NB,
                                            esw_id * OVS_DOCA_MAX_MIRRORS_PER_ESW,
                                            OVS_DOCA_MAX_MIRRORS_PER_ESW);
    if (!doca_mirrors->id_pool) {
        goto err_create_id_pool;
    }

    doca_mirrors->rfm = refmap_create("doca-mirrors", sizeof(struct doca_mirror_key),
                                      sizeof(struct doca_mirror_ctx),
                                      doca_mirror_ctx_init,
                                      doca_mirror_ctx_uninit,
                                      doca_mirror_ctx_dump,
                                      false);
    if (!doca_mirrors->rfm) {
        goto err_create_refmap;
    }

    return doca_mirrors;

err_create_refmap:
    id_fpool_destroy(doca_mirrors->id_pool);
err_create_id_pool:
    free(doca_mirrors);
    return NULL;
}

void
doca_mirrors_destroy(struct doca_mirrors *doca_mirrors)
{
    if (!doca_mirrors) {
        return;
    }

    refmap_destroy(doca_mirrors->rfm);
    id_fpool_destroy(doca_mirrors->id_pool);
    free(doca_mirrors);
}

struct doca_mirror_ctx *
doca_mirror_ctx_ref(struct doca_mirrors *doca_mirrors,
                    struct doca_flow_resource_mirror_cfg *cfg)
{
    struct doca_mirror_key key = {
        .nr_targets = cfg->nr_targets,
    };
    struct doca_mirror_arg arg = {
        .doca_mirrors = doca_mirrors,
        .key = &key,
    };

    if (cfg->nr_targets == 0 || cfg->nr_targets > DOCA_MIRRORS_MAX_TARGETS ||
        cfg->fwd.type != DOCA_FLOW_FWD_NONE) {
        return NULL;
    }

    memcpy(&key.targets, cfg->target, sizeof(*cfg->target) * cfg->nr_targets);
    return refmap_ref(doca_mirrors->rfm, &key, &arg);
}

void
doca_mirror_ctx_unref(struct doca_mirror_ctx *mirror_ctx)
{
    if (!mirror_ctx) {
        return;
    }

    refmap_unref(mirror_ctx->doca_mirrors->rfm, mirror_ctx);
}

doca_error_t
doca_mirror_ctx_bind(struct doca_mirror_ctx *mirror_ctx, void *bindable_object,
                     uint32_t *shared_mirror_id)
{
    struct doca_mirror_key *key = refmap_key_from_value(mirror_ctx->doca_mirrors->rfm, mirror_ctx);
    struct doca_flow_shared_resource_cfg resource_cfg;
    unsigned int tid = netdev_offload_thread_id();
    struct doca_mirror_binding *binding;
    doca_error_t err;

    ovs_mutex_lock(&bound_mirrors_lock);
    LIST_FOR_EACH (binding, mirror_ctx_list_node, &mirror_ctx->binding_list) {
        if (binding->bindable_object == bindable_object) {
            *shared_mirror_id = binding->shared_mirror_id;
            ovs_mutex_unlock(&bound_mirrors_lock);
            return 0;
        }
    }
    ovs_mutex_unlock(&bound_mirrors_lock);

    binding = xzalloc(sizeof *binding);
    if (!id_fpool_new_id(mirror_ctx->doca_mirrors->id_pool, tid, &binding->shared_mirror_id)) {
        err = DOCA_ERROR_FULL;
        goto err_new_id;
    }

    resource_cfg = (struct doca_flow_shared_resource_cfg) {
        .mirror_cfg.nr_targets = key->nr_targets,
        .mirror_cfg.target = key->targets,
    };

    err = doca_flow_shared_resource_set_cfg(DOCA_FLOW_SHARED_RESOURCE_MIRROR,
                                            binding->shared_mirror_id, &resource_cfg);
    if (err != DOCA_SUCCESS) {
        VLOG_ERR("Failed to set resource_cfg of shared mirror (id %d), err %d - %s",
                 binding->shared_mirror_id, err, doca_error_get_descr(err));
        goto err_set_cfg;
    }

    err = doca_flow_shared_resources_bind(DOCA_FLOW_SHARED_RESOURCE_MIRROR,
                                          &binding->shared_mirror_id, 1, bindable_object);
    if (err) {
        goto err_set_cfg;
    }

    binding->mirror_ctx = mirror_ctx;
    binding->bindable_object = bindable_object;

    /* DOCA doesn't release bound shared resources till the bindable object (pipe/port)
     * is destroyed. So keep ID live till object release event.
     */
    refmap_ref_value(mirror_ctx->doca_mirrors->rfm, mirror_ctx, true);

    ovs_mutex_lock(&bound_mirrors_lock);
    ovs_list_push_back(&mirror_ctx->binding_list, &binding->mirror_ctx_list_node);
    ovs_list_push_back(&mirrors_binding_list, &binding->mirrors_binding_list_node);
    ovs_mutex_unlock(&bound_mirrors_lock);

    atomic_count_inc(&mirrors_count);

    *shared_mirror_id = binding->shared_mirror_id;
    return 0;

err_set_cfg:
    id_fpool_free_id(mirror_ctx->doca_mirrors->id_pool, tid, binding->shared_mirror_id);
err_new_id:
    free(binding);
    return err;
}

void
doca_mirrors_bindable_object_released(void *bindable_object)
{
    unsigned int tid = netdev_offload_thread_id();
    struct doca_mirror_binding *binding;
    struct doca_mirror_ctx *mirror_ctx;

    if (!atomic_count_get(&mirrors_count)) {
        return;
    }

    ovs_mutex_lock(&bound_mirrors_lock);
    LIST_FOR_EACH_SAFE (binding, mirrors_binding_list_node, &mirrors_binding_list) {
        if (binding->bindable_object != bindable_object) {
            continue;
        }

        mirror_ctx = binding->mirror_ctx;
        ovs_list_remove(&binding->mirrors_binding_list_node);
        ovs_list_remove(&binding->mirror_ctx_list_node);
        id_fpool_free_id(mirror_ctx->doca_mirrors->id_pool, tid, binding->shared_mirror_id);
        doca_mirror_ctx_unref(mirror_ctx);
        free(binding);

        atomic_count_dec(&mirrors_count);
    }
    ovs_mutex_unlock(&bound_mirrors_lock);
}
